"""
Errors
======

"""

import ipaddress

from typing import Optional, Union


class GeoIP2Error(RuntimeError):
    """There was a generic error in GeoIP2.

    This class represents a generic error. It extends :py:exc:`RuntimeError`
    and does not add any additional attributes.

    """


class AddressNotFoundError(GeoIP2Error):
    """The address you were looking up was not found.

    .. attribute:: ip_address

      The IP address used in the lookup. This is only available for database
      lookups.

      :type: str

    .. attribute:: network

      The network associated with the error. In particular, this is the
      largest network where no address would be found. This is only
      available for database lookups.

      :type: ipaddress.IPv4Network or ipaddress.IPv6Network

    """

    ip_address: Optional[str]
    _prefix_len: Optional[int]

    def __init__(
        self,
        message: str,
        ip_address: Optional[str] = None,
        prefix_len: Optional[int] = None,
    ) -> None:
        super().__init__(message)
        self.ip_address = ip_address
        self._prefix_len = prefix_len

    @property
    def network(self) -> Optional[Union[ipaddress.IPv4Network, ipaddress.IPv6Network]]:
        """The network for the error"""

        if self.ip_address is None or self._prefix_len is None:
            return None
        return ipaddress.ip_network(f"{self.ip_address}/{self._prefix_len}", False)


class AuthenticationError(GeoIP2Error):
    """There was a problem authenticating the request."""


class HTTPError(GeoIP2Error):
    """There was an error when making your HTTP request.

    This class represents an HTTP transport error. It extends
    :py:exc:`GeoIP2Error` and adds attributes of its own.

    :ivar http_status: The HTTP status code returned
    :ivar uri: The URI queried
    :ivar decoded_content: The decoded response content

    """

    def __init__(
        self,
        message: str,
        http_status: Optional[int] = None,
        uri: Optional[str] = None,
        decoded_content: Optional[str] = None,
    ) -> None:
        super().__init__(message)
        self.http_status = http_status
        self.uri = uri
        self.decoded_content = decoded_content


class InvalidRequestError(GeoIP2Error):
    """The request was invalid."""


class OutOfQueriesError(GeoIP2Error):
    """Your account is out of funds for the service queried."""


class PermissionRequiredError(GeoIP2Error):
    """Your account does not have permission to access this service."""



#!/usr/bin/python
# -*- coding: utf-8 -*-

from __future__ import print_function

import argparse
import geoip2.database
import random
import socket
import struct
import timeit

parser = argparse.ArgumentParser(description="Benchmark maxminddb.")
parser.add_argument("--count", default=250000, type=int, help="number of lookups")
parser.add_argument("--mode", default=0, type=int, help="reader mode to use")
parser.add_argument("--file", default="GeoIP2-City.mmdb", help="path to mmdb file")

args = parser.parse_args()

reader = geoip2.database.Reader(args.file, mode=args.mode)


def lookup_ip_address():
    ip = socket.inet_ntoa(struct.pack("!L", random.getrandbits(32)))
    try:
        record = reader.city(str(ip))
    except geoip2.errors.AddressNotFoundError:
        pass


elapsed = timeit.timeit(
    "lookup_ip_address()",
    setup="from __main__ import lookup_ip_address",
    number=args.count,
)

print(args.count / elapsed, "lookups per second")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import asyncio
import copy
import ipaddress
import json
import sys
from typing import cast, Dict
import unittest

sys.path.append("..")

# httpretty currently doesn't work, but mocket with the compat interface
# does.
from mocket import Mocket  # type: ignore
from mocket.plugins.httpretty import httpretty, httprettified  # type: ignore
import geoip2
from geoip2.errors import (
    AddressNotFoundError,
    AuthenticationError,
    GeoIP2Error,
    HTTPError,
    InvalidRequestError,
    OutOfQueriesError,
    PermissionRequiredError,
)
from geoip2.webservice import AsyncClient, Client


class TestBaseClient(unittest.TestCase):
    base_uri = "https://geoip.maxmind.com/geoip/v2.1/"
    country = {
        "continent": {"code": "NA", "geoname_id": 42, "names": {"en": "North America"}},
        "country": {
            "geoname_id": 1,
            "iso_code": "US",
            "names": {"en": "United States of America"},
        },
        "maxmind": {"queries_remaining": 11},
        "registered_country": {
            "geoname_id": 2,
            "is_in_european_union": True,
            "iso_code": "DE",
            "names": {"en": "Germany"},
        },
        "traits": {
            "ip_address": "1.2.3.4",
            "is_anycast": True,
            "network": "1.2.3.0/24",
        },
    }

    # this is not a comprehensive representation of the
    # JSON from the server
    insights = cast(Dict, copy.deepcopy(country))
    insights["traits"]["user_count"] = 2
    insights["traits"]["static_ip_score"] = 1.3

    def _content_type(self, endpoint):
        return (
            "application/vnd.maxmind.com-"
            + endpoint
            + "+json; charset=UTF-8; version=1.0"
        )

    @httprettified
    def test_country_ok(self):
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "country/1.2.3.4",
            body=json.dumps(self.country),
            status=200,
            content_type=self._content_type("country"),
        )
        country = self.run_client(self.client.country("1.2.3.4"))
        self.assertEqual(
            type(country), geoip2.models.Country, "return value of client.country"
        )
        self.assertEqual(country.continent.geoname_id, 42, "continent geoname_id is 42")
        self.assertEqual(country.continent.code, "NA", "continent code is NA")
        self.assertEqual(
            country.continent.name, "North America", "continent name is North America"
        )
        self.assertEqual(country.country.geoname_id, 1, "country geoname_id is 1")
        self.assertIs(
            country.country.is_in_european_union,
            False,
            "country is_in_european_union is False",
        )
        self.assertEqual(country.country.iso_code, "US", "country iso_code is US")
        self.assertEqual(
            country.country.names, {"en": "United States of America"}, "country names"
        )
        self.assertEqual(
            country.country.name,
            "United States of America",
            "country name is United States of America",
        )
        self.assertEqual(
            country.maxmind.queries_remaining, 11, "queries_remaining is 11"
        )
        self.assertIs(
            country.registered_country.is_in_european_union,
            True,
            "registered_country is_in_european_union is True",
        )
        self.assertEqual(
            country.traits.network, ipaddress.ip_network("1.2.3.0/24"), "network"
        )
        self.assertTrue(country.traits.is_anycast)
        self.assertEqual(country.raw, self.country, "raw response is correct")

    @httprettified
    def test_me(self):
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "country/me",
            body=json.dumps(self.country),
            status=200,
            content_type=self._content_type("country"),
        )
        implicit_me = self.run_client(self.client.country())
        self.assertEqual(
            type(implicit_me), geoip2.models.Country, "country() returns Country object"
        )
        explicit_me = self.run_client(self.client.country())
        self.assertEqual(
            type(explicit_me),
            geoip2.models.Country,
            "country('me') returns Country object",
        )

    @httprettified
    def test_200_error(self):
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "country/1.1.1.1",
            body="",
            status=200,
            content_type=self._content_type("country"),
        )
        with self.assertRaisesRegex(
            GeoIP2Error, "could not decode the response as JSON"
        ):
            self.run_client(self.client.country("1.1.1.1"))

    @httprettified
    def test_bad_ip_address(self):
        with self.assertRaisesRegex(
            ValueError, "'1.2.3' does not appear to be an IPv4 " "or IPv6 address"
        ):
            self.run_client(self.client.country("1.2.3"))

    @httprettified
    def test_no_body_error(self):
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "country/" + "1.2.3.7",
            body="",
            status=400,
            content_type=self._content_type("country"),
        )
        with self.assertRaisesRegex(
            HTTPError, "Received a 400 error for .* with no body"
        ):
            self.run_client(self.client.country("1.2.3.7"))

    @httprettified
    def test_weird_body_error(self):
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "country/" + "1.2.3.8",
            body='{"wierd": 42}',
            status=400,
            content_type=self._content_type("country"),
        )
        with self.assertRaisesRegex(
            HTTPError,
            "Response contains JSON but it does not " "specify code or error keys",
        ):
            self.run_client(self.client.country("1.2.3.8"))

    @httprettified
    def test_bad_body_error(self):
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "country/" + "1.2.3.9",
            body="bad body",
            status=400,
            content_type=self._content_type("country"),
        )
        with self.assertRaisesRegex(
            HTTPError, "it did not include the expected JSON body"
        ):
            self.run_client(self.client.country("1.2.3.9"))

    @httprettified
    def test_500_error(self):
        httpretty.register_uri(
            httpretty.GET, self.base_uri + "country/" + "1.2.3.10", status=500
        )
        with self.assertRaisesRegex(HTTPError, r"Received a server error \(500\) for"):
            self.run_client(self.client.country("1.2.3.10"))

    @httprettified
    def test_300_error(self):
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "country/" + "1.2.3.11",
            status=300,
            content_type=self._content_type("country"),
        )
        with self.assertRaisesRegex(
            HTTPError, r"Received a very surprising HTTP status \(300\) for"
        ):
            self.run_client(self.client.country("1.2.3.11"))

    @httprettified
    def test_ip_address_required(self):
        self._test_error(400, "IP_ADDRESS_REQUIRED", InvalidRequestError)

    @httprettified
    def test_ip_address_not_found(self):
        self._test_error(404, "IP_ADDRESS_NOT_FOUND", AddressNotFoundError)

    @httprettified
    def test_ip_address_reserved(self):
        self._test_error(400, "IP_ADDRESS_RESERVED", AddressNotFoundError)

    @httprettified
    def test_permission_required(self):
        self._test_error(403, "PERMISSION_REQUIRED", PermissionRequiredError)

    @httprettified
    def test_auth_invalid(self):
        self._test_error(400, "AUTHORIZATION_INVALID", AuthenticationError)

    @httprettified
    def test_license_key_required(self):
        self._test_error(401, "LICENSE_KEY_REQUIRED", AuthenticationError)

    @httprettified
    def test_account_id_required(self):
        self._test_error(401, "ACCOUNT_ID_REQUIRED", AuthenticationError)

    @httprettified
    def test_user_id_required(self):
        self._test_error(401, "USER_ID_REQUIRED", AuthenticationError)

    @httprettified
    def test_account_id_unkown(self):
        self._test_error(401, "ACCOUNT_ID_UNKNOWN", AuthenticationError)

    @httprettified
    def test_user_id_unkown(self):
        self._test_error(401, "USER_ID_UNKNOWN", AuthenticationError)

    @httprettified
    def test_out_of_queries_error(self):
        self._test_error(402, "OUT_OF_QUERIES", OutOfQueriesError)

    def _test_error(self, status, error_code, error_class):
        msg = "Some error message"
        body = {"error": msg, "code": error_code}
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "country/1.2.3.18",
            body=json.dumps(body),
            status=status,
            content_type=self._content_type("country"),
        )
        with self.assertRaisesRegex(error_class, msg):
            self.run_client(self.client.country("1.2.3.18"))

    @httprettified
    def test_unknown_error(self):
        msg = "Unknown error type"
        ip = "1.2.3.19"
        body = {"error": msg, "code": "UNKNOWN_TYPE"}
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "country/" + ip,
            body=json.dumps(body),
            status=400,
            content_type=self._content_type("country"),
        )
        with self.assertRaisesRegex(InvalidRequestError, msg):
            self.run_client(self.client.country(ip))

    @httprettified
    def test_request(self):
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "country/" + "1.2.3.4",
            body=json.dumps(self.country),
            status=200,
            content_type=self._content_type("country"),
        )
        self.run_client(self.client.country("1.2.3.4"))
        request = httpretty.last_request

        self.assertEqual(
            request.path, "/geoip/v2.1/country/1.2.3.4", "correct URI is used"
        )

        # This is to prevent breakage if header normalization in Mocket
        # changes again in the future.
        headers = {k.lower(): v for k, v in request.headers.items()}
        self.assertEqual(headers["accept"], "application/json", "correct Accept header")
        self.assertRegex(
            headers["user-agent"],
            "^GeoIP2-Python-Client/",
            "Correct User-Agent",
        )
        self.assertEqual(
            headers["authorization"],
            "Basic NDI6YWJjZGVmMTIzNDU2",
            "correct auth",
        )

    @httprettified
    def test_city_ok(self):
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "city/" + "1.2.3.4",
            body=json.dumps(self.country),
            status=200,
            content_type=self._content_type("city"),
        )
        city = self.run_client(self.client.city("1.2.3.4"))
        self.assertEqual(type(city), geoip2.models.City, "return value of client.city")
        self.assertEqual(
            city.traits.network, ipaddress.ip_network("1.2.3.0/24"), "network"
        )
        self.assertTrue(city.traits.is_anycast)

    @httprettified
    def test_insights_ok(self):
        httpretty.register_uri(
            httpretty.GET,
            self.base_uri + "insights/1.2.3.4",
            body=json.dumps(self.insights),
            status=200,
            content_type=self._content_type("country"),
        )
        insights = self.run_client(self.client.insights("1.2.3.4"))
        self.assertEqual(
            type(insights), geoip2.models.Insights, "return value of client.insights"
        )
        self.assertEqual(
            insights.traits.network, ipaddress.ip_network("1.2.3.0/24"), "network"
        )
        self.assertTrue(insights.traits.is_anycast)
        self.assertEqual(insights.traits.static_ip_score, 1.3, "static_ip_score is 1.3")
        self.assertEqual(insights.traits.user_count, 2, "user_count is 2")

    def test_named_constructor_args(self):
        id = 47
        key = "1234567890ab"
        client = self.client_class(account_id=id, license_key=key)
        self.assertEqual(client._account_id, str(id))
        self.assertEqual(client._license_key, key)

    def test_missing_constructor_args(self):
        with self.assertRaises(TypeError):
            self.client_class(license_key="1234567890ab")

        with self.assertRaises(TypeError):
            self.client_class("47")


class TestClient(TestBaseClient):
    def setUp(self):
        self.client_class = Client
        self.client = Client(42, "abcdef123456")

    def run_client(self, v):
        return v


class TestAsyncClient(TestBaseClient):
    def setUp(self):
        self._loop = asyncio.new_event_loop()
        self.client_class = AsyncClient
        self.client = AsyncClient(42, "abcdef123456")

    def tearDown(self):
        self._loop.run_until_complete(self.client.close())
        self._loop.close()

    def run_client(self, v):
        return self._loop.run_until_complete(v)


del TestBaseClient


if __name__ == "__main__":
    unittest.main()



"""

Records
=======

"""
# pylint:disable=too-many-arguments,too-many-instance-attributes,too-many-locals

import ipaddress

# pylint:disable=R0903
from abc import ABCMeta
from typing import Dict, List, Optional, Type, Union

from geoip2.mixins import SimpleEquality


class Record(SimpleEquality, metaclass=ABCMeta):
    """All records are subclasses of the abstract class ``Record``."""

    def __repr__(self) -> str:
        args = ", ".join(f"{k}={v!r}" for k, v in self.__dict__.items())
        return f"{self.__module__}.{self.__class__.__name__}({args})"


class PlaceRecord(Record, metaclass=ABCMeta):
    """All records with :py:attr:`names` subclass :py:class:`PlaceRecord`."""

    names: Dict[str, str]
    _locales: List[str]

    def __init__(
        self,
        locales: Optional[List[str]] = None,
        names: Optional[Dict[str, str]] = None,
    ) -> None:
        if locales is None:
            locales = ["en"]
        self._locales = locales
        if names is None:
            names = {}
        self.names = names

    @property
    def name(self) -> Optional[str]:
        """Dict with locale codes as keys and localized name as value."""
        # pylint:disable=E1101
        return next((self.names.get(x) for x in self._locales if x in self.names), None)


class City(PlaceRecord):
    """Contains data for the city record associated with an IP address.

    This class contains the city-level data associated with an IP address.

    This record is returned by ``city``, ``enterprise``, and ``insights``.

    Attributes:

    .. attribute:: confidence

      A value from 0-100 indicating MaxMind's
      confidence that the city is correct. This attribute is only available
      from the Insights end point and the Enterprise database.

      :type: int

    .. attribute:: geoname_id

      The GeoName ID for the city.

      :type: int

    .. attribute:: name

      The name of the city based on the locales list passed to the
      constructor.

      :type: str

    .. attribute:: names

      A dictionary where the keys are locale codes
      and the values are names.

      :type: dict

    """

    confidence: Optional[int]
    geoname_id: Optional[int]

    def __init__(
        self,
        locales: Optional[List[str]] = None,
        confidence: Optional[int] = None,
        geoname_id: Optional[int] = None,
        names: Optional[Dict[str, str]] = None,
        **_,
    ) -> None:
        self.confidence = confidence
        self.geoname_id = geoname_id
        super().__init__(locales, names)


class Continent(PlaceRecord):
    """Contains data for the continent record associated with an IP address.

    This class contains the continent-level data associated with an IP
    address.

    Attributes:


    .. attribute:: code

      A two character continent code like "NA" (North America)
      or "OC" (Oceania).

      :type: str

    .. attribute:: geoname_id

      The GeoName ID for the continent.

      :type: int

    .. attribute:: name

      Returns the name of the continent based on the locales list passed to
      the constructor.

      :type: str

    .. attribute:: names

      A dictionary where the keys are locale codes
      and the values are names.

      :type: dict

    """

    code: Optional[str]
    geoname_id: Optional[int]

    def __init__(
        self,
        locales: Optional[List[str]] = None,
        code: Optional[str] = None,
        geoname_id: Optional[int] = None,
        names: Optional[Dict[str, str]] = None,
        **_,
    ) -> None:
        self.code = code
        self.geoname_id = geoname_id
        super().__init__(locales, names)


class Country(PlaceRecord):
    """Contains data for the country record associated with an IP address.

    This class contains the country-level data associated with an IP address.

    Attributes:


    .. attribute:: confidence

      A value from 0-100 indicating MaxMind's confidence that
      the country is correct. This attribute is only available from the
      Insights end point and the Enterprise database.

      :type: int

    .. attribute:: geoname_id

      The GeoName ID for the country.

      :type: int

    .. attribute:: is_in_european_union

      This is true if the country is a member state of the European Union.

      :type: bool

    .. attribute:: iso_code

      The two-character `ISO 3166-1
      <https://en.wikipedia.org/wiki/ISO_3166-1>`_ alpha code for the
      country.

      :type: str

    .. attribute:: name

      The name of the country based on the locales list passed to the
      constructor.

      :type: str

    .. attribute:: names

      A dictionary where the keys are locale codes and the values
      are names.

      :type: dict

    """

    confidence: Optional[int]
    geoname_id: Optional[int]
    is_in_european_union: bool
    iso_code: Optional[str]

    def __init__(
        self,
        locales: Optional[List[str]] = None,
        confidence: Optional[int] = None,
        geoname_id: Optional[int] = None,
        is_in_european_union: bool = False,
        iso_code: Optional[str] = None,
        names: Optional[Dict[str, str]] = None,
        **_,
    ) -> None:
        self.confidence = confidence
        self.geoname_id = geoname_id
        self.is_in_european_union = is_in_european_union
        self.iso_code = iso_code
        super().__init__(locales, names)


class RepresentedCountry(Country):
    """Contains data for the represented country associated with an IP address.

    This class contains the country-level data associated with an IP address
    for the IP's represented country. The represented country is the country
    represented by something like a military base.

    Attributes:


    .. attribute:: confidence

      A value from 0-100 indicating MaxMind's confidence that
      the country is correct. This attribute is only available from the
      Insights end point and the Enterprise database.

      :type: int

    .. attribute:: geoname_id

      The GeoName ID for the country.

      :type: int

    .. attribute:: is_in_european_union

      This is true if the country is a member state of the European Union.

      :type: bool

    .. attribute:: iso_code

      The two-character `ISO 3166-1
      <https://en.wikipedia.org/wiki/ISO_3166-1>`_ alpha code for the country.

      :type: str

    .. attribute:: name

      The name of the country based on the locales list passed to the
      constructor.

      :type: str

    .. attribute:: names

      A dictionary where the keys are locale codes and the values
      are names.

      :type: dict


    .. attribute:: type

      A string indicating the type of entity that is representing the
      country. Currently we only return ``military`` but this could expand to
      include other types in the future.

      :type: str

    """

    type: Optional[str]

    def __init__(
        self,
        locales: Optional[List[str]] = None,
        confidence: Optional[int] = None,
        geoname_id: Optional[int] = None,
        is_in_european_union: bool = False,
        iso_code: Optional[str] = None,
        names: Optional[Dict[str, str]] = None,
        # pylint:disable=redefined-builtin
        type: Optional[str] = None,
        **_,
    ) -> None:
        self.type = type
        super().__init__(
            locales, confidence, geoname_id, is_in_european_union, iso_code, names
        )


class Location(Record):
    """Contains data for the location record associated with an IP address.

    This class contains the location data associated with an IP address.

    This record is returned by ``city``, ``enterprise``, and ``insights``.

    Attributes:

    .. attribute:: average_income

      The average income in US dollars associated with the requested IP
      address. This attribute is only available from the Insights end point.

      :type: int

    .. attribute:: accuracy_radius

      The approximate accuracy radius in kilometers around the latitude and
      longitude for the IP address. This is the radius where we have a 67%
      confidence that the device using the IP address resides within the
      circle centered at the latitude and longitude with the provided radius.

      :type: int

    .. attribute:: latitude

      The approximate latitude of the location associated with the IP
      address. This value is not precise and should not be used to identify a
      particular address or household.

      :type: float

    .. attribute:: longitude

      The approximate longitude of the location associated with the IP
      address. This value is not precise and should not be used to identify a
      particular address or household.

      :type: float

    .. attribute:: metro_code

      The metro code of the location if the
      location is in the US. MaxMind returns the same metro codes as the
      `Google AdWords API
      <https://developers.google.com/adwords/api/docs/appendix/cities-DMAregions>`_.

      :type: int

      .. attribute:: population_density

      The estimated population per square kilometer associated with the IP
      address. This attribute is only available from the Insights end point.

      :type: int

    .. attribute:: time_zone

      The time zone associated with location, as specified by the `IANA Time
      Zone Database <https://www.iana.org/time-zones>`_, e.g.,
      "America/New_York".

      :type: str

    """

    average_income: Optional[int]
    accuracy_radius: Optional[int]
    latitude: Optional[float]
    longitude: Optional[float]
    metro_code: Optional[int]
    population_density: Optional[int]
    time_zone: Optional[str]

    def __init__(
        self,
        average_income: Optional[int] = None,
        accuracy_radius: Optional[int] = None,
        latitude: Optional[float] = None,
        longitude: Optional[float] = None,
        metro_code: Optional[int] = None,
        population_density: Optional[int] = None,
        time_zone: Optional[str] = None,
        **_,
    ) -> None:
        self.average_income = average_income
        self.accuracy_radius = accuracy_radius
        self.latitude = latitude
        self.longitude = longitude
        self.metro_code = metro_code
        self.population_density = population_density
        self.time_zone = time_zone


class MaxMind(Record):
    """Contains data related to your MaxMind account.

    Attributes:

    .. attribute:: queries_remaining

      The number of remaining queries you have
      for the end point you are calling.

      :type: int

    """

    queries_remaining: Optional[int]

    def __init__(self, queries_remaining: Optional[int] = None, **_) -> None:
        self.queries_remaining = queries_remaining


class Postal(Record):
    """Contains data for the postal record associated with an IP address.

    This class contains the postal data associated with an IP address.

    This attribute is returned by ``city``, ``enterprise``, and ``insights``.

    Attributes:

    .. attribute:: code

      The postal code of the location. Postal
      codes are not available for all countries. In some countries, this will
      only contain part of the postal code.

      :type: str

    .. attribute:: confidence

      A value from 0-100 indicating
      MaxMind's confidence that the postal code is correct. This attribute is
      only available from the Insights end point and the Enterprise database.

      :type: int

    """

    code: Optional[str]
    confidence: Optional[int]

    def __init__(
        self, code: Optional[str] = None, confidence: Optional[int] = None, **_
    ) -> None:
        self.code = code
        self.confidence = confidence


class Subdivision(PlaceRecord):
    """Contains data for the subdivisions associated with an IP address.

    This class contains the subdivision data associated with an IP address.

    This attribute is returned by ``city``, ``enterprise``, and ``insights``.

    Attributes:

    .. attribute:: confidence

      This is a value from 0-100 indicating MaxMind's
      confidence that the subdivision is correct. This attribute is only
      available from the Insights end point and the Enterprise database.

      :type: int

    .. attribute:: geoname_id

      This is a GeoName ID for the subdivision.

      :type: int

    .. attribute:: iso_code

      This is a string up to three characters long
      contain the subdivision portion of the `ISO 3166-2 code
      <https://en.wikipedia.org/wiki/ISO_3166-2>`_.

      :type: str

    .. attribute:: name

      The name of the subdivision based on the locales list passed to the
      constructor.

      :type: str

    .. attribute:: names

      A dictionary where the keys are locale codes and the
      values are names

      :type: dict

    """

    confidence: Optional[int]
    geoname_id: Optional[int]
    iso_code: Optional[str]

    def __init__(
        self,
        locales: Optional[List[str]] = None,
        confidence: Optional[int] = None,
        geoname_id: Optional[int] = None,
        iso_code: Optional[str] = None,
        names: Optional[Dict[str, str]] = None,
        **_,
    ) -> None:
        self.confidence = confidence
        self.geoname_id = geoname_id
        self.iso_code = iso_code
        super().__init__(locales, names)


class Subdivisions(tuple):
    """A tuple-like collection of subdivisions associated with an IP address.

    This class contains the subdivisions of the country associated with the
    IP address from largest to smallest.

    For instance, the response for Oxford in the United Kingdom would have
    England as the first element and Oxfordshire as the second element.

    This attribute is returned by ``city``, ``enterprise``, and ``insights``.
    """

    def __new__(
        cls: Type["Subdivisions"], locales: Optional[List[str]], *subdivisions
    ) -> "Subdivisions":
        subobjs = tuple(Subdivision(locales, **x) for x in subdivisions)
        obj = super().__new__(cls, subobjs)  # type: ignore
        return obj

    def __init__(
        self, locales: Optional[List[str]], *subdivisions  # pylint:disable=W0613
    ) -> None:
        self._locales = locales
        super().__init__()

    @property
    def most_specific(self) -> Subdivision:
        """The most specific (smallest) subdivision available.

        If there are no :py:class:`Subdivision` objects for the response,
        this returns an empty :py:class:`Subdivision`.

        :type: :py:class:`Subdivision`
        """
        try:
            return self[-1]
        except IndexError:
            return Subdivision(self._locales)


class Traits(Record):
    """Contains data for the traits record associated with an IP address.

    This class contains the traits data associated with an IP address.

    This class has the following attributes:


    .. attribute:: autonomous_system_number

      The `autonomous system
      number <https://en.wikipedia.org/wiki/Autonomous_system_(Internet)>`_
      associated with the IP address. This attribute is only available from
      the City Plus and Insights web services and the Enterprise database.

      :type: int

    .. attribute:: autonomous_system_organization

      The organization associated with the registered `autonomous system
      number <https://en.wikipedia.org/wiki/Autonomous_system_(Internet)>`_ for
      the IP address. This attribute is only available from the City Plus and
      Insights web service end points and the Enterprise database.

      :type: str

    .. attribute:: connection_type

      The connection type may take the following values:

      - Dialup
      - Cable/DSL
      - Corporate
      - Cellular
      - Satellite

      Additional values may be added in the future.

      This attribute is only available from the City Plus and Insights web
      service end points and the Enterprise database.

      :type: str

    .. attribute:: domain

      The second level domain associated with the
      IP address. This will be something like "example.com" or
      "example.co.uk", not "foo.example.com". This attribute is only available
      from the City Plus and Insights web service end points and the
      Enterprise database.

      :type: str

    .. attribute:: ip_address

      The IP address that the data in the model
      is for. If you performed a "me" lookup against the web service, this
      will be the externally routable IP address for the system the code is
      running on. If the system is behind a NAT, this may differ from the IP
      address locally assigned to it.

      :type: str

    .. attribute:: is_anonymous

      This is true if the IP address belongs to any sort of anonymous network.
      This attribute is only available from Insights.

      :type: bool

    .. attribute:: is_anonymous_proxy

      This is true if the IP is an anonymous proxy.

      :type: bool

      .. deprecated:: 2.2.0
        Use our our `GeoIP2 Anonymous IP database
        <https://www.maxmind.com/en/geoip2-anonymous-ip-database GeoIP2>`_
        instead.

    .. attribute:: is_anonymous_vpn

      This is true if the IP address is registered to an anonymous VPN
      provider.

      If a VPN provider does not register subnets under names associated with
      them, we will likely only flag their IP ranges using the
      ``is_hosting_provider`` attribute.

      This attribute is only available from Insights.

      :type: bool

    .. attribute:: is_anycast

      This returns true if the IP address belongs to an
      `anycast network <https://en.wikipedia.org/wiki/Anycast>`_.
      This is available for the GeoIP2 Country, City Plus, and Insights
      web services and the GeoIP2 Country, City, and Enterprise databases.

      :type: bool

    .. attribute:: is_hosting_provider

      This is true if the IP address belongs to a hosting or VPN provider
      (see description of ``is_anonymous_vpn`` attribute).
      This attribute is only available from Insights.

      :type: bool

    .. attribute:: is_legitimate_proxy

      This attribute is true if MaxMind believes this IP address to be a
      legitimate proxy, such as an internal VPN used by a corporation. This
      attribute is only available in the Enterprise database.

      :type: bool

    .. attribute:: is_public_proxy

      This is true if the IP address belongs to a public proxy. This attribute
      is only available from Insights.

      :type: bool

    .. attribute:: is_residential_proxy

      This is true if the IP address is on a suspected anonymizing network
      and belongs to a residential ISP. This attribute is only available from
      Insights.

      :type: bool


    .. attribute:: is_satellite_provider

      This is true if the IP address is from a satellite provider that
      provides service to multiple countries.

      :type: bool

      .. deprecated:: 2.2.0
        Due to the increased coverage by mobile carriers, very few
        satellite providers now serve multiple countries. As a result, the
        output does not provide sufficiently relevant data for us to maintain
        it.

    .. attribute:: is_tor_exit_node

      This is true if the IP address is a Tor exit node. This attribute is
      only available from Insights.

      :type: bool

    .. attribute:: isp

      The name of the ISP associated with the IP address. This attribute is
      only available from the City Plus and Insights web services and the
      Enterprise database.

      :type: str

    .. attribute: mobile_country_code

      The `mobile country code (MCC)
      <https://en.wikipedia.org/wiki/Mobile_country_code>`_ associated with the
      IP address and ISP. This attribute is available from the City Plus and
      Insights web services and the Enterprise database.

      :type: str

    .. attribute: mobile_network_code

      The `mobile network code (MNC)
      <https://en.wikipedia.org/wiki/Mobile_country_code>`_ associated with the
      IP address and ISP. This attribute is available from the City Plus and
      Insights web services and the Enterprise database.

      :type: str

    .. attribute:: network

      The network associated with the record. In particular, this is the
      largest network where all of the fields besides ip_address have the same
      value.

      :type: ipaddress.IPv4Network or ipaddress.IPv6Network

    .. attribute:: organization

      The name of the organization associated with the IP address. This
      attribute is only available from the City Plus and Insights web services
      and the Enterprise database.

      :type: str

    .. attribute:: static_ip_score

      An indicator of how static or dynamic an IP address is. The value ranges
      from 0 to 99.99 with higher values meaning a greater static association.
      For example, many IP addresses with a user_type of cellular have a
      lifetime under one. Static Cable/DSL IPs typically have a lifetime above
      thirty.

      This indicator can be useful for deciding whether an IP address represents
      the same user over time. This attribute is only available from
      Insights.

      :type: float

    .. attribute:: user_count

      The estimated number of users sharing the IP/network during the past 24
      hours. For IPv4, the count is for the individual IP. For IPv6, the count
      is for the /64 network. This attribute is only available from
      Insights.

      :type: int

    .. attribute:: user_type

      The user type associated with the IP
      address. This can be one of the following values:

      * business
      * cafe
      * cellular
      * college
      * consumer_privacy_network
      * content_delivery_network
      * dialup
      * government
      * hosting
      * library
      * military
      * residential
      * router
      * school
      * search_engine_spider
      * traveler

      This attribute is only available from the Insights end point and the
      Enterprise database.

      :type: str

    """

    autonomous_system_number: Optional[int]
    autonomous_system_organization: Optional[str]
    connection_type: Optional[str]
    domain: Optional[str]
    ip_address: Optional[str]
    is_anonymous: bool
    is_anonymous_proxy: bool
    is_anonymous_vpn: bool
    is_anycast: bool
    is_hosting_provider: bool
    is_legitimate_proxy: bool
    is_public_proxy: bool
    is_residential_proxy: bool
    is_satellite_provider: bool
    is_tor_exit_node: bool
    isp: Optional[str]
    mobile_country_code: Optional[str]
    mobile_network_code: Optional[str]
    organization: Optional[str]
    static_ip_score: Optional[float]
    user_count: Optional[int]
    user_type: Optional[str]
    _network: Optional[Union[ipaddress.IPv4Network, ipaddress.IPv6Network]]
    _prefix_len: Optional[int]

    def __init__(
        self,
        autonomous_system_number: Optional[int] = None,
        autonomous_system_organization: Optional[str] = None,
        connection_type: Optional[str] = None,
        domain: Optional[str] = None,
        is_anonymous: bool = False,
        is_anonymous_proxy: bool = False,
        is_anonymous_vpn: bool = False,
        is_hosting_provider: bool = False,
        is_legitimate_proxy: bool = False,
        is_public_proxy: bool = False,
        is_residential_proxy: bool = False,
        is_satellite_provider: bool = False,
        is_tor_exit_node: bool = False,
        isp: Optional[str] = None,
        ip_address: Optional[str] = None,
        network: Optional[str] = None,
        organization: Optional[str] = None,
        prefix_len: Optional[int] = None,
        static_ip_score: Optional[float] = None,
        user_count: Optional[int] = None,
        user_type: Optional[str] = None,
        mobile_country_code: Optional[str] = None,
        mobile_network_code: Optional[str] = None,
        is_anycast: bool = False,
        **_,
    ) -> None:
        self.autonomous_system_number = autonomous_system_number
        self.autonomous_system_organization = autonomous_system_organization
        self.connection_type = connection_type
        self.domain = domain
        self.is_anonymous = is_anonymous
        self.is_anonymous_proxy = is_anonymous_proxy
        self.is_anonymous_vpn = is_anonymous_vpn
        self.is_anycast = is_anycast
        self.is_hosting_provider = is_hosting_provider
        self.is_legitimate_proxy = is_legitimate_proxy
        self.is_public_proxy = is_public_proxy
        self.is_residential_proxy = is_residential_proxy
        self.is_satellite_provider = is_satellite_provider
        self.is_tor_exit_node = is_tor_exit_node
        self.isp = isp
        self.mobile_country_code = mobile_country_code
        self.mobile_network_code = mobile_network_code
        self.organization = organization
        self.static_ip_score = static_ip_score
        self.user_type = user_type
        self.user_count = user_count
        self.ip_address = ip_address
        if network is None:
            self._network = None
        else:
            self._network = ipaddress.ip_network(network, False)
        # We don't construct the network using prefix_len here as that is
        # for database lookups. Customers using the database tend to be
        # much more performance sensitive than web service users.
        self._prefix_len = prefix_len

    @property
    def network(self) -> Optional[Union[ipaddress.IPv4Network, ipaddress.IPv6Network]]:
        """The network for the record"""
        # This code is duplicated for performance reasons
        network = self._network
        if network is not None:
            return network

        ip_address = self.ip_address
        prefix_len = self._prefix_len
        if ip_address is None or prefix_len is None:
            return None
        network = ipaddress.ip_network(f"{ip_address}/{prefix_len}", False)
        self._network = network
        return network



"""
============================
WebServices Client API
============================

This class provides a client API for all the GeoIP2 web services. The web
services are Country, City Plus, and Insights. Each service returns a
different set of data about an IP address, with Country returning the least
data and Insights the most.

Each service is represented by a different model class, and these model
classes in turn contain multiple record classes. The record classes have
attributes which contain data about the IP address.

If the service does not return a particular piece of data for an IP address,
the associated attribute is not populated.

The service may not return any information for an entire record, in which
case all of the attributes for that record class will be empty.

SSL
---

Requests to the web service are always made with SSL.

"""

import ipaddress
import json
from typing import Any, Dict, cast, List, Optional, Type, Union

import aiohttp
import aiohttp.http
import requests
import requests.utils

import geoip2
import geoip2.models
from geoip2.errors import (
    AddressNotFoundError,
    AuthenticationError,
    GeoIP2Error,
    HTTPError,
    InvalidRequestError,
    OutOfQueriesError,
    PermissionRequiredError,
)
from geoip2.models import City, Country, Insights
from geoip2.types import IPAddress

_AIOHTTP_UA = (
    f"GeoIP2-Python-Client/{geoip2.__version__} {aiohttp.http.SERVER_SOFTWARE}"
)

_REQUEST_UA = (
    f"GeoIP2-Python-Client/{geoip2.__version__} {requests.utils.default_user_agent()}"
)


class BaseClient:  # pylint: disable=missing-class-docstring, too-few-public-methods
    _account_id: str
    _host: str
    _license_key: str
    _locales: List[str]
    _timeout: float

    def __init__(
        self,
        account_id: int,
        license_key: str,
        host: str,
        locales: Optional[List[str]],
        timeout: float,
    ) -> None:
        """Construct a Client."""
        # pylint: disable=too-many-arguments
        if locales is None:
            locales = ["en"]

        self._locales = locales
        # requests 2.12.2 requires that the username passed to auth be bytes
        # or a string, with the former being preferred.
        self._account_id = (
            account_id if isinstance(account_id, bytes) else str(account_id)
        )
        self._license_key = license_key
        self._base_uri = f"https://{host}/geoip/v2.1"
        self._timeout = timeout

    def _uri(self, path: str, ip_address: IPAddress) -> str:
        if ip_address != "me":
            ip_address = ipaddress.ip_address(ip_address)
        return "/".join([self._base_uri, path, str(ip_address)])

    @staticmethod
    def _handle_success(body: str, uri: str) -> Any:
        try:
            return json.loads(body)
        except ValueError as ex:
            raise GeoIP2Error(
                f"Received a 200 response for {uri}"
                " but could not decode the response as "
                "JSON: " + ", ".join(ex.args),
                200,
                uri,
            ) from ex

    def _exception_for_error(
        self, status: int, content_type: str, body: str, uri: str
    ) -> GeoIP2Error:
        if 400 <= status < 500:
            return self._exception_for_4xx_status(status, content_type, body, uri)
        if 500 <= status < 600:
            return self._exception_for_5xx_status(status, uri, body)
        return self._exception_for_non_200_status(status, uri, body)

    def _exception_for_4xx_status(
        self, status: int, content_type: str, body: str, uri: str
    ) -> GeoIP2Error:
        if not body:
            return HTTPError(
                f"Received a {status} error for {uri} with no body.",
                status,
                uri,
                body,
            )
        if content_type.find("json") == -1:
            return HTTPError(
                f"Received a {status} for {uri} with the following body: {body}",
                status,
                uri,
                body,
            )
        try:
            decoded_body = json.loads(body)
        except ValueError as ex:
            return HTTPError(
                f"Received a {status} error for {uri} but it did not include "
                + "the expected JSON body: "
                + ", ".join(ex.args),
                status,
                uri,
                body,
            )

        if "code" in decoded_body and "error" in decoded_body:
            return self._exception_for_web_service_error(
                decoded_body.get("error"), decoded_body.get("code"), status, uri
            )
        return HTTPError(
            "Response contains JSON but it does not specify code or error keys",
            status,
            uri,
            body,
        )

    @staticmethod
    def _exception_for_web_service_error(
        message: str, code: str, status: int, uri: str
    ) -> Union[
        AuthenticationError,
        AddressNotFoundError,
        PermissionRequiredError,
        OutOfQueriesError,
        InvalidRequestError,
    ]:
        if code in ("IP_ADDRESS_NOT_FOUND", "IP_ADDRESS_RESERVED"):
            return AddressNotFoundError(message)
        if code in (
            "ACCOUNT_ID_REQUIRED",
            "ACCOUNT_ID_UNKNOWN",
            "AUTHORIZATION_INVALID",
            "LICENSE_KEY_REQUIRED",
            "USER_ID_REQUIRED",
            "USER_ID_UNKNOWN",
        ):
            return AuthenticationError(message)
        if code in ("INSUFFICIENT_FUNDS", "OUT_OF_QUERIES"):
            return OutOfQueriesError(message)
        if code == "PERMISSION_REQUIRED":
            return PermissionRequiredError(message)

        return InvalidRequestError(message, code, status, uri)

    @staticmethod
    def _exception_for_5xx_status(
        status: int, uri: str, body: Optional[str]
    ) -> HTTPError:
        return HTTPError(
            f"Received a server error ({status}) for {uri}",
            status,
            uri,
            body,
        )

    @staticmethod
    def _exception_for_non_200_status(
        status: int, uri: str, body: Optional[str]
    ) -> HTTPError:
        return HTTPError(
            f"Received a very surprising HTTP status ({status}) for {uri}",
            status,
            uri,
            body,
        )


class AsyncClient(BaseClient):
    """An async GeoIP2 client.

    It accepts the following required arguments:

    :param account_id: Your MaxMind account ID.
    :param license_key: Your MaxMind license key.

    Go to https://www.maxmind.com/en/my_license_key to see your MaxMind
    account ID and license key.

    The following keyword arguments are also accepted:

    :param host: The hostname to make a request against. This defaults to
      "geoip.maxmind.com". To use the GeoLite2 web service instead of the
      GeoIP2 web service, set this to "geolite.info". To use the Sandbox
      GeoIP2 web service instead of the production GeoIP2 web service, set
      this to "sandbox.maxmind.com". The sandbox allows you to experiment
      with the API without affecting your production data.
    :param locales: This is list of locale codes. This argument will be
      passed on to record classes to use when their name properties are
      called. The default value is ['en'].

      The order of the locales is significant. When a record class has
      multiple names (country, city, etc.), its name property will return
      the name in the first locale that has one.

      Note that the only locale which is always present in the GeoIP2
      data is "en". If you do not include this locale, the name property
      may end up returning None even when the record has an English name.

      Currently, the valid locale codes are:

      * de -- German
      * en -- English names may still include accented characters if that is
        the accepted spelling in English. In other words, English does not
        mean ASCII.
      * es -- Spanish
      * fr -- French
      * ja -- Japanese
      * pt-BR -- Brazilian Portuguese
      * ru -- Russian
      * zh-CN -- Simplified Chinese.
    :param timeout: The timeout in seconds to use when waiting on the request.
      This sets both the connect timeout and the read timeout. The default is
      60.
    :param proxy: The URL of an HTTP proxy to use. It may optionally include
      a basic auth username and password, e.g.,
      ``http://username:password@host:port``.

    """

    _existing_session: aiohttp.ClientSession
    _proxy: Optional[str]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        account_id: int,
        license_key: str,
        host: str = "geoip.maxmind.com",
        locales: Optional[List[str]] = None,
        timeout: float = 60,
        proxy: Optional[str] = None,
    ) -> None:
        super().__init__(
            account_id,
            license_key,
            host,
            locales,
            timeout,
        )
        self._proxy = proxy

    async def city(self, ip_address: IPAddress = "me") -> City:
        """Call City Plus endpoint with the specified IP.

        :param ip_address: IPv4 or IPv6 address as a string. If no
           address is provided, the address that the web service is
           called from will be used.

        :returns: :py:class:`geoip2.models.City` object

        """
        return cast(
            City, await self._response_for("city", geoip2.models.City, ip_address)
        )

    async def country(self, ip_address: IPAddress = "me") -> Country:
        """Call the GeoIP2 Country endpoint with the specified IP.

        :param ip_address: IPv4 or IPv6 address as a string. If no address
          is provided, the address that the web service is called from will
          be used.

        :returns: :py:class:`geoip2.models.Country` object

        """
        return cast(
            Country,
            await self._response_for("country", geoip2.models.Country, ip_address),
        )

    async def insights(self, ip_address: IPAddress = "me") -> Insights:
        """Call the Insights endpoint with the specified IP.

        Insights is only supported by the GeoIP2 web service. The GeoLite2 web
        service does not support it.

        :param ip_address: IPv4 or IPv6 address as a string. If no address
          is provided, the address that the web service is called from will
          be used.

        :returns: :py:class:`geoip2.models.Insights` object

        """
        return cast(
            Insights,
            await self._response_for("insights", geoip2.models.Insights, ip_address),
        )

    async def _session(self) -> aiohttp.ClientSession:
        if not hasattr(self, "_existing_session"):
            self._existing_session = aiohttp.ClientSession(
                auth=aiohttp.BasicAuth(self._account_id, self._license_key),
                headers={"Accept": "application/json", "User-Agent": _AIOHTTP_UA},
                timeout=aiohttp.ClientTimeout(total=self._timeout),
            )

        return self._existing_session

    async def _response_for(
        self,
        path: str,
        model_class: Union[Type[Insights], Type[City], Type[Country]],
        ip_address: IPAddress,
    ) -> Union[Country, City, Insights]:
        uri = self._uri(path, ip_address)
        session = await self._session()
        async with await session.get(uri, proxy=self._proxy) as response:
            status = response.status
            content_type = response.content_type
            body = await response.text()
            if status != 200:
                raise self._exception_for_error(status, content_type, body, uri)
            decoded_body = self._handle_success(body, uri)
            return model_class(decoded_body, locales=self._locales)

    async def close(self):
        """Close underlying session

        This will close the session and any associated connections.
        """
        if hasattr(self, "_existing_session"):
            await self._existing_session.close()

    async def __aenter__(self) -> "AsyncClient":
        return self

    async def __aexit__(self, exc_type: None, exc_value: None, traceback: None) -> None:
        await self.close()


class Client(BaseClient):
    """A synchronous GeoIP2 client.

    It accepts the following required arguments:

    :param account_id: Your MaxMind account ID.
    :param license_key: Your MaxMind license key.

    Go to https://www.maxmind.com/en/my_license_key to see your MaxMind
    account ID and license key.

    The following keyword arguments are also accepted:

    :param host: The hostname to make a request against. This defaults to
      "geoip.maxmind.com". To use the GeoLite2 web service instead of the
      GeoIP2 web service, set this to "geolite.info". To use the Sandbox
      GeoIP2 web service instead of the production GeoIP2 web service, set
      this to "sandbox.maxmind.com". The sandbox allows you to experiment
      with the API without affecting your production data.
    :param locales: This is list of locale codes. This argument will be
      passed on to record classes to use when their name properties are
      called. The default value is ['en'].

      The order of the locales is significant. When a record class has
      multiple names (country, city, etc.), its name property will return
      the name in the first locale that has one.

      Note that the only locale which is always present in the GeoIP2
      data is "en". If you do not include this locale, the name property
      may end up returning None even when the record has an English name.

      Currently, the valid locale codes are:

      * de -- German
      * en -- English names may still include accented characters if that is
        the accepted spelling in English. In other words, English does not
        mean ASCII.
      * es -- Spanish
      * fr -- French
      * ja -- Japanese
      * pt-BR -- Brazilian Portuguese
      * ru -- Russian
      * zh-CN -- Simplified Chinese.
    :param timeout: The timeout in seconds to use when waiting on the request.
      This sets both the connect timeout and the read timeout. The default is
      60.
    :param proxy: The URL of an HTTP proxy to use. It may optionally include
      a basic auth username and password, e.g.,
      ``http://username:password@host:port``.


    """

    _session: requests.Session
    _proxies: Optional[Dict[str, str]]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        account_id: int,
        license_key: str,
        host: str = "geoip.maxmind.com",
        locales: Optional[List[str]] = None,
        timeout: float = 60,
        proxy: Optional[str] = None,
    ) -> None:
        super().__init__(account_id, license_key, host, locales, timeout)
        self._session = requests.Session()
        self._session.auth = (self._account_id, self._license_key)
        self._session.headers["Accept"] = "application/json"
        self._session.headers["User-Agent"] = _REQUEST_UA
        if proxy is None:
            self._proxies = None
        else:
            self._proxies = {"https": proxy}

    def city(self, ip_address: IPAddress = "me") -> City:
        """Call City Plus endpoint with the specified IP.

        :param ip_address: IPv4 or IPv6 address as a string. If no
           address is provided, the address that the web service is
           called from will be used.

        :returns: :py:class:`geoip2.models.City` object

        """
        return cast(City, self._response_for("city", geoip2.models.City, ip_address))

    def country(self, ip_address: IPAddress = "me") -> Country:
        """Call the GeoIP2 Country endpoint with the specified IP.

        :param ip_address: IPv4 or IPv6 address as a string. If no address
          is provided, the address that the web service is called from will
          be used.

        :returns: :py:class:`geoip2.models.Country` object

        """
        return cast(
            Country, self._response_for("country", geoip2.models.Country, ip_address)
        )

    def insights(self, ip_address: IPAddress = "me") -> Insights:
        """Call the Insights endpoint with the specified IP.

        Insights is only supported by the GeoIP2 web service. The GeoLite2 web
        service does not support it.

        :param ip_address: IPv4 or IPv6 address as a string. If no address
          is provided, the address that the web service is called from will
          be used.

        :returns: :py:class:`geoip2.models.Insights` object

        """
        return cast(
            Insights, self._response_for("insights", geoip2.models.Insights, ip_address)
        )

    def _response_for(
        self,
        path: str,
        model_class: Union[Type[Insights], Type[City], Type[Country]],
        ip_address: IPAddress,
    ) -> Union[Country, City, Insights]:
        uri = self._uri(path, ip_address)
        response = self._session.get(uri, proxies=self._proxies, timeout=self._timeout)
        status = response.status_code
        content_type = response.headers["Content-Type"]
        body = response.text
        if status != 200:
            raise self._exception_for_error(status, content_type, body, uri)
        decoded_body = self._handle_success(body, uri)
        return model_class(decoded_body, locales=self._locales)

    def close(self):
        """Close underlying session

        This will close the session and any associated connections.
        """
        self._session.close()

    def __enter__(self) -> "Client":
        return self

    def __exit__(self, exc_type: None, exc_value: None, traceback: None) -> None:
        self.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import ipaddress
import sys
import unittest

sys.path.append("..")

import geoip2.database
import geoip2.errors
import maxminddb

try:
    import maxminddb.extension
except ImportError:
    maxminddb.extension = None  # type: ignore


class BaseTestReader(unittest.TestCase):
    def test_language_list(self) -> None:
        reader = geoip2.database.Reader(
            "tests/data/test-data/GeoIP2-Country-Test.mmdb",
            ["xx", "ru", "pt-BR", "es", "en"],
        )
        record = reader.country("81.2.69.160")

        self.assertEqual(record.country.name, "Великобритания")
        reader.close()

    def test_unknown_address(self) -> None:
        reader = geoip2.database.Reader("tests/data/test-data/GeoIP2-City-Test.mmdb")
        with self.assertRaisesRegex(
            geoip2.errors.AddressNotFoundError,
            "The address 10.10.10.10 is not in the " "database.",
        ):
            reader.city("10.10.10.10")
        reader.close()

    def test_unknown_address_network(self) -> None:
        reader = geoip2.database.Reader("tests/data/test-data/GeoIP2-City-Test.mmdb")
        try:
            reader.city("10.10.10.10")
            self.fail("Expected AddressNotFoundError")
        except geoip2.errors.AddressNotFoundError as e:
            self.assertEqual(e.network, ipaddress.ip_network("10.0.0.0/8"))
        except Exception as e:
            self.fail(f"Expected AddressNotFoundError, got {type(e)}: {str(e)}")
        finally:
            reader.close()

    def test_wrong_database(self) -> None:
        reader = geoip2.database.Reader("tests/data/test-data/GeoIP2-City-Test.mmdb")
        with self.assertRaisesRegex(
            TypeError,
            "The country method cannot be used with " "the GeoIP2-City database",
        ):
            reader.country("1.1.1.1")
        reader.close()

    def test_invalid_address(self) -> None:
        reader = geoip2.database.Reader("tests/data/test-data/GeoIP2-City-Test.mmdb")
        with self.assertRaisesRegex(
            ValueError, "u?'invalid' does not appear to be an " "IPv4 or IPv6 address"
        ):
            reader.city("invalid")
        reader.close()

    def test_anonymous_ip(self) -> None:
        reader = geoip2.database.Reader(
            "tests/data/test-data/GeoIP2-Anonymous-IP-Test.mmdb"
        )
        ip_address = "1.2.0.1"

        record = reader.anonymous_ip(ip_address)
        self.assertEqual(record.is_anonymous, True)
        self.assertEqual(record.is_anonymous_vpn, True)
        self.assertEqual(record.is_hosting_provider, False)
        self.assertEqual(record.is_public_proxy, False)
        self.assertEqual(record.is_residential_proxy, False)
        self.assertEqual(record.is_tor_exit_node, False)
        self.assertEqual(record.ip_address, ip_address)
        self.assertEqual(record.network, ipaddress.ip_network("1.2.0.0/16"))
        reader.close()

    def test_anonymous_ip_all_set(self) -> None:
        reader = geoip2.database.Reader(
            "tests/data/test-data/GeoIP2-Anonymous-IP-Test.mmdb"
        )
        ip_address = "81.2.69.1"

        record = reader.anonymous_ip(ip_address)
        self.assertEqual(record.is_anonymous, True)
        self.assertEqual(record.is_anonymous_vpn, True)
        self.assertEqual(record.is_hosting_provider, True)
        self.assertEqual(record.is_public_proxy, True)
        self.assertEqual(record.is_residential_proxy, True)
        self.assertEqual(record.is_tor_exit_node, True)
        self.assertEqual(record.ip_address, ip_address)
        self.assertEqual(record.network, ipaddress.ip_network("81.2.69.0/24"))
        reader.close()

    def test_asn(self) -> None:
        reader = geoip2.database.Reader("tests/data/test-data/GeoLite2-ASN-Test.mmdb")

        ip_address = "1.128.0.0"
        record = reader.asn(ip_address)

        self.assertEqual(record, eval(repr(record)), "ASN repr can be eval'd")

        self.assertEqual(record.autonomous_system_number, 1221)
        self.assertEqual(record.autonomous_system_organization, "Telstra Pty Ltd")
        self.assertEqual(record.ip_address, ip_address)
        self.assertEqual(record.network, ipaddress.ip_network("1.128.0.0/11"))

        self.assertRegex(
            str(record),
            r"geoip2.models.ASN\(.*1\.128\.0\.0.*\)",
            "str representation is correct",
        )

        reader.close()

    def test_city(self) -> None:
        reader = geoip2.database.Reader("tests/data/test-data/GeoIP2-City-Test.mmdb")
        record = reader.city("81.2.69.160")

        self.assertEqual(
            record.country.name, "United Kingdom", "The default locale is en"
        )
        self.assertEqual(record.country.is_in_european_union, False)
        self.assertEqual(
            record.location.accuracy_radius, 100, "The accuracy_radius is populated"
        )
        self.assertEqual(record.registered_country.is_in_european_union, False)
        self.assertFalse(record.traits.is_anycast)

        record = reader.city("214.1.1.0")
        self.assertTrue(record.traits.is_anycast)

        reader.close()

    def test_connection_type(self) -> None:
        reader = geoip2.database.Reader(
            "tests/data/test-data/GeoIP2-Connection-Type-Test.mmdb"
        )
        ip_address = "1.0.1.0"

        record = reader.connection_type(ip_address)

        self.assertEqual(
            record, eval(repr(record)), "ConnectionType repr can be eval'd"
        )

        self.assertEqual(record.connection_type, "Cellular")
        self.assertEqual(record.ip_address, ip_address)
        self.assertEqual(record.network, ipaddress.ip_network("1.0.1.0/24"))

        self.assertRegex(
            str(record),
            r"ConnectionType\(\{.*Cellular.*\}\)",
            "ConnectionType str representation is reasonable",
        )

        reader.close()

    def test_country(self) -> None:
        reader = geoip2.database.Reader("tests/data/test-data/GeoIP2-Country-Test.mmdb")
        record = reader.country("81.2.69.160")
        self.assertEqual(
            record.traits.ip_address, "81.2.69.160", "IP address is added to model"
        )
        self.assertEqual(record.traits.network, ipaddress.ip_network("81.2.69.160/27"))
        self.assertEqual(record.country.is_in_european_union, False)
        self.assertEqual(record.registered_country.is_in_european_union, False)
        self.assertFalse(record.traits.is_anycast)

        record = reader.country("214.1.1.0")
        self.assertTrue(record.traits.is_anycast)

        reader.close()

    def test_domain(self) -> None:
        reader = geoip2.database.Reader("tests/data/test-data/GeoIP2-Domain-Test.mmdb")

        ip_address = "1.2.0.0"
        record = reader.domain(ip_address)

        self.assertEqual(record, eval(repr(record)), "Domain repr can be eval'd")

        self.assertEqual(record.domain, "maxmind.com")
        self.assertEqual(record.ip_address, ip_address)
        self.assertEqual(record.network, ipaddress.ip_network("1.2.0.0/16"))

        self.assertRegex(
            str(record),
            r"Domain\(\{.*maxmind.com.*\}\)",
            "Domain str representation is reasonable",
        )

        reader.close()

    def test_enterprise(self) -> None:
        with geoip2.database.Reader(
            "tests/data/test-data/GeoIP2-Enterprise-Test.mmdb"
        ) as reader:
            ip_address = "74.209.24.0"
            record = reader.enterprise(ip_address)
            self.assertEqual(record.city.confidence, 11)
            self.assertEqual(record.country.confidence, 99)
            self.assertEqual(record.country.geoname_id, 6252001)
            self.assertEqual(record.country.is_in_european_union, False)
            self.assertEqual(record.location.accuracy_radius, 27)
            self.assertEqual(record.registered_country.is_in_european_union, False)
            self.assertEqual(record.traits.connection_type, "Cable/DSL")
            self.assertTrue(record.traits.is_legitimate_proxy)
            self.assertEqual(record.traits.ip_address, ip_address)
            self.assertEqual(
                record.traits.network, ipaddress.ip_network("74.209.16.0/20")
            )
            self.assertFalse(record.traits.is_anycast)

            record = reader.enterprise("149.101.100.0")
            self.assertEqual(record.traits.mobile_country_code, "310")
            self.assertEqual(record.traits.mobile_network_code, "004")

            record = reader.enterprise("214.1.1.0")
            self.assertTrue(record.traits.is_anycast)

    def test_isp(self) -> None:
        with geoip2.database.Reader(
            "tests/data/test-data/GeoIP2-ISP-Test.mmdb"
        ) as reader:
            ip_address = "1.128.0.0"
            record = reader.isp(ip_address)
            self.assertEqual(record, eval(repr(record)), "ISP repr can be eval'd")

            self.assertEqual(record.autonomous_system_number, 1221)
            self.assertEqual(record.autonomous_system_organization, "Telstra Pty Ltd")
            self.assertEqual(record.isp, "Telstra Internet")
            self.assertEqual(record.organization, "Telstra Internet")
            self.assertEqual(record.ip_address, ip_address)
            self.assertEqual(record.network, ipaddress.ip_network("1.128.0.0/11"))

            self.assertRegex(
                str(record),
                r"ISP\(\{.*Telstra.*\}\)",
                "ISP str representation is reasonable",
            )

            record = reader.isp("149.101.100.0")

            self.assertEqual(record.mobile_country_code, "310")
            self.assertEqual(record.mobile_network_code, "004")

    def test_context_manager(self) -> None:
        with geoip2.database.Reader(
            "tests/data/test-data/GeoIP2-Country-Test.mmdb"
        ) as reader:
            record = reader.country("81.2.69.160")
            self.assertEqual(record.traits.ip_address, "81.2.69.160")


@unittest.skipUnless(maxminddb.extension, "No C extension module found. Skipping tests")
class TestExtensionReader(BaseTestReader):
    mode = geoip2.database.MODE_MMAP_EXT


class TestMMAPReader(BaseTestReader):
    mode = geoip2.database.MODE_MMAP


class TestFileReader(BaseTestReader):
    mode = geoip2.database.MODE_FILE


class TestMemoryReader(BaseTestReader):
    mode = geoip2.database.MODE_MEMORY


class TestFDReader(unittest.TestCase):
    mode = geoip2.database.MODE_FD


class TestAutoReader(BaseTestReader):
    mode = geoip2.database.MODE_AUTO



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import sys
from typing import Dict
import unittest

sys.path.append("..")

import geoip2.models


class TestModels(unittest.TestCase):
    def test_insights_full(self) -> None:
        raw = {
            "city": {
                "confidence": 76,
                "geoname_id": 9876,
                "names": {"en": "Minneapolis"},
            },
            "continent": {
                "code": "NA",
                "geoname_id": 42,
                "names": {"en": "North America"},
            },
            "country": {
                "confidence": 99,
                "geoname_id": 1,
                "iso_code": "US",
                "names": {"en": "United States of America"},
            },
            "location": {
                "average_income": 24626,
                "accuracy_radius": 1500,
                "latitude": 44.98,
                "longitude": 93.2636,
                "metro_code": 765,
                "population_density": 1341,
                "time_zone": "America/Chicago",
            },
            "postal": {
                "code": "55401",
                "confidence": 33,
            },
            "subdivisions": [
                {
                    "confidence": 88,
                    "geoname_id": 574635,
                    "iso_code": "MN",
                    "names": {"en": "Minnesota"},
                },
                {
                    "geoname_id": 123,
                    "iso_code": "HP",
                    "names": {"en": "Hennepin"},
                },
            ],
            "registered_country": {
                "geoname_id": 2,
                "iso_code": "CA",
                "names": {"en": "Canada"},
            },
            "represented_country": {
                "geoname_id": 3,
                "is_in_european_union": True,
                "iso_code": "GB",
                "names": {"en": "United Kingdom"},
                "type": "military",
            },
            "traits": {
                "autonomous_system_number": 1234,
                "autonomous_system_organization": "AS Organization",
                "connection_type": "Cable/DSL",
                "domain": "example.com",
                "ip_address": "1.2.3.4",
                "is_anonymous": True,
                "is_anonymous_proxy": True,
                "is_anonymous_vpn": True,
                "is_anycast": True,
                "is_hosting_provider": True,
                "is_public_proxy": True,
                "is_residential_proxy": True,
                "is_satellite_provider": True,
                "is_tor_exit_node": True,
                "isp": "Comcast",
                "network_speed": "cable/DSL",
                "organization": "Blorg",
                "static_ip_score": 1.3,
                "user_count": 2,
                "user_type": "college",
            },
        }

        model = geoip2.models.Insights(raw)
        self.assertEqual(
            type(model), geoip2.models.Insights, "geoip2.models.Insights object"
        )
        self.assertEqual(
            type(model.city), geoip2.records.City, "geoip2.records.City object"
        )
        self.assertEqual(
            type(model.continent),
            geoip2.records.Continent,
            "geoip2.records.Continent object",
        )
        self.assertEqual(
            type(model.country), geoip2.records.Country, "geoip2.records.Country object"
        )
        self.assertEqual(
            type(model.registered_country),
            geoip2.records.Country,
            "geoip2.records.Country object",
        )
        self.assertEqual(
            type(model.represented_country),
            geoip2.records.RepresentedCountry,
            "geoip2.records.RepresentedCountry object",
        )
        self.assertEqual(
            type(model.location),
            geoip2.records.Location,
            "geoip2.records.Location object",
        )
        self.assertEqual(
            type(model.subdivisions[0]),
            geoip2.records.Subdivision,
            "geoip2.records.Subdivision object",
        )
        self.assertEqual(
            type(model.traits), geoip2.records.Traits, "geoip2.records.Traits object"
        )
        self.assertEqual(model.raw, raw, "raw method returns raw input")
        self.assertEqual(
            model.subdivisions[0].iso_code, "MN", "div 1 has correct iso_code"
        )
        self.assertEqual(
            model.subdivisions[0].confidence, 88, "div 1 has correct confidence"
        )
        self.assertEqual(
            model.subdivisions[0].geoname_id, 574635, "div 1 has correct geoname_id"
        )
        self.assertEqual(
            model.subdivisions[0].names, {"en": "Minnesota"}, "div 1 names are correct"
        )
        self.assertEqual(
            model.subdivisions[1].name, "Hennepin", "div 2 has correct name"
        )
        self.assertEqual(
            model.subdivisions.most_specific.iso_code,
            "HP",
            "subdivisions.most_specific returns HP",
        )
        self.assertEqual(
            model.represented_country.name,
            "United Kingdom",
            "represented_country name is correct",
        )
        self.assertEqual(
            model.represented_country.type,
            "military",
            "represented_country type is correct",
        )
        self.assertEqual(model.location.average_income, 24626, "correct average_income")
        self.assertEqual(model.location.latitude, 44.98, "correct latitude")
        self.assertEqual(model.location.longitude, 93.2636, "correct longitude")
        self.assertEqual(model.location.metro_code, 765, "correct metro_code")
        self.assertEqual(
            model.location.population_density, 1341, "correct population_density"
        )

        self.assertRegex(
            str(model),
            r"^geoip2.models.Insights\(\{.*geoname_id.*\}, \[.*en.*\]\)",
            "Insights str representation looks reasonable",
        )

        self.assertEqual(model, eval(repr(model)), "Insights repr can be eval'd")

        self.assertRegex(
            str(model.location),
            r"^geoip2.records.Location\(.*longitude=.*\)",
            "Location str representation is reasonable",
        )

        self.assertEqual(
            model.location, eval(repr(model.location)), "Location repr can be eval'd"
        )

        self.assertIs(model.country.is_in_european_union, False)
        self.assertIs(model.registered_country.is_in_european_union, False)
        self.assertIs(model.represented_country.is_in_european_union, True)

        self.assertIs(model.traits.is_anonymous, True)
        self.assertIs(model.traits.is_anonymous_proxy, True)
        self.assertIs(model.traits.is_anonymous_vpn, True)
        self.assertIs(model.traits.is_anycast, True)
        self.assertIs(model.traits.is_hosting_provider, True)
        self.assertIs(model.traits.is_public_proxy, True)
        self.assertIs(model.traits.is_residential_proxy, True)
        self.assertIs(model.traits.is_satellite_provider, True)
        self.assertIs(model.traits.is_tor_exit_node, True)
        self.assertEqual(model.traits.user_count, 2)
        self.assertEqual(model.traits.static_ip_score, 1.3)

    def test_insights_min(self) -> None:
        model = geoip2.models.Insights({"traits": {"ip_address": "5.6.7.8"}})
        self.assertEqual(
            type(model), geoip2.models.Insights, "geoip2.models.Insights object"
        )
        self.assertEqual(
            type(model.city), geoip2.records.City, "geoip2.records.City object"
        )
        self.assertEqual(
            type(model.continent),
            geoip2.records.Continent,
            "geoip2.records.Continent object",
        )
        self.assertEqual(
            type(model.country), geoip2.records.Country, "geoip2.records.Country object"
        )
        self.assertEqual(
            type(model.registered_country),
            geoip2.records.Country,
            "geoip2.records.Country object",
        )
        self.assertEqual(
            type(model.location),
            geoip2.records.Location,
            "geoip2.records.Location object",
        )
        self.assertEqual(
            type(model.traits), geoip2.records.Traits, "geoip2.records.Traits object"
        )
        self.assertEqual(
            type(model.subdivisions.most_specific),
            geoip2.records.Subdivision,
            "geoip2.records.Subdivision object returned even"
            "when none are available.",
        )
        self.assertEqual(
            model.subdivisions.most_specific.names, {}, "Empty names hash returned"
        )

    def test_city_full(self) -> None:
        raw = {
            "continent": {
                "code": "NA",
                "geoname_id": 42,
                "names": {"en": "North America"},
            },
            "country": {
                "geoname_id": 1,
                "iso_code": "US",
                "names": {"en": "United States of America"},
            },
            "registered_country": {
                "geoname_id": 2,
                "iso_code": "CA",
                "names": {"en": "Canada"},
            },
            "traits": {
                "ip_address": "1.2.3.4",
                "is_satellite_provider": True,
            },
        }
        model = geoip2.models.City(raw)
        self.assertEqual(type(model), geoip2.models.City, "geoip2.models.City object")
        self.assertEqual(
            type(model.city), geoip2.records.City, "geoip2.records.City object"
        )
        self.assertEqual(
            type(model.continent),
            geoip2.records.Continent,
            "geoip2.records.Continent object",
        )
        self.assertEqual(
            type(model.country), geoip2.records.Country, "geoip2.records.Country object"
        )
        self.assertEqual(
            type(model.registered_country),
            geoip2.records.Country,
            "geoip2.records.Country object",
        )
        self.assertEqual(
            type(model.location),
            geoip2.records.Location,
            "geoip2.records.Location object",
        )
        self.assertEqual(
            type(model.traits), geoip2.records.Traits, "geoip2.records.Traits object"
        )
        self.assertEqual(model.raw, raw, "raw method returns raw input")
        self.assertEqual(model.continent.geoname_id, 42, "continent geoname_id is 42")
        self.assertEqual(model.continent.code, "NA", "continent code is NA")
        self.assertEqual(
            model.continent.names, {"en": "North America"}, "continent names is correct"
        )
        self.assertEqual(
            model.continent.name, "North America", "continent name is correct"
        )
        self.assertEqual(model.country.geoname_id, 1, "country geoname_id is 1")
        self.assertEqual(model.country.iso_code, "US", "country iso_code is US")
        self.assertEqual(
            model.country.names,
            {"en": "United States of America"},
            "country names is correct",
        )
        self.assertEqual(
            model.country.name, "United States of America", "country name is correct"
        )
        self.assertEqual(model.country.confidence, None, "country confidence is None")
        self.assertEqual(
            model.registered_country.iso_code, "CA", "registered_country iso_code is CA"
        )
        self.assertEqual(
            model.registered_country.names,
            {"en": "Canada"},
            "registered_country names is correct",
        )
        self.assertEqual(
            model.registered_country.name,
            "Canada",
            "registered_country name is correct",
        )
        self.assertEqual(
            model.traits.is_anonymous_proxy,
            False,
            "traits is_anonymous_proxy returns False by default",
        )
        self.assertEqual(
            model.traits.is_anycast,
            False,
            "traits is_anycast returns False by default",
        )
        self.assertEqual(
            model.traits.is_satellite_provider,
            True,
            "traits is_setellite_provider is True",
        )
        self.assertEqual(model.raw, raw, "raw method produces raw output")

        self.assertRegex(
            str(model), r"^geoip2.models.City\(\{.*geoname_id.*\}, \[.*en.*\]\)"
        )

        self.assertFalse(model == True, "__eq__ does not blow up on weird input")

    def test_unknown_keys(self) -> None:
        model = geoip2.models.City(
            {
                "city": {"invalid": 0},
                "continent": {
                    "invalid": 0,
                    "names": {"invalid": 0},
                },
                "country": {
                    "invalid": 0,
                    "names": {"invalid": 0},
                },
                "location": {"invalid": 0},
                "postal": {"invalid": 0},
                "subdivisions": [
                    {
                        "invalid": 0,
                        "names": {
                            "invalid": 0,
                        },
                    },
                ],
                "registered_country": {
                    "invalid": 0,
                    "names": {
                        "invalid": 0,
                    },
                },
                "represented_country": {
                    "invalid": 0,
                    "names": {
                        "invalid": 0,
                    },
                },
                "traits": {"ip_address": "1.2.3.4", "invalid": "blah"},
                "unk_base": {"blah": 1},
            }
        )
        with self.assertRaises(AttributeError):
            model.unk_base  # type: ignore
        with self.assertRaises(AttributeError):
            model.traits.invalid  # type: ignore
        self.assertEqual(model.traits.ip_address, "1.2.3.4", "correct ip")


class TestNames(unittest.TestCase):
    raw: Dict = {
        "continent": {
            "code": "NA",
            "geoname_id": 42,
            "names": {
                "de": "Nordamerika",
                "en": "North America",
                "es": "América del Norte",
                "ja": "北アメリカ",
                "pt-BR": "América do Norte",
                "ru": "Северная Америка",
                "zh-CN": "北美洲",
            },
        },
        "country": {
            "geoname_id": 1,
            "iso_code": "US",
            "names": {
                "en": "United States of America",
                "fr": "États-Unis",
                "zh-CN": "美国",
            },
        },
        "traits": {
            "ip_address": "1.2.3.4",
        },
    }

    def test_names(self) -> None:
        model = geoip2.models.Country(self.raw, locales=["sq", "ar"])
        self.assertEqual(
            model.continent.names,
            self.raw["continent"]["names"],
            "Correct names dict for continent",
        )
        self.assertEqual(
            model.country.names,
            self.raw["country"]["names"],
            "Correct names dict for country",
        )

    def test_three_locales(self) -> None:
        model = geoip2.models.Country(self.raw, locales=["fr", "zh-CN", "en"])
        self.assertEqual(
            model.continent.name,
            "北美洲",
            "continent name is in Chinese (no French available)",
        )
        self.assertEqual(model.country.name, "États-Unis", "country name is in French")

    def test_two_locales(self) -> None:
        model = geoip2.models.Country(self.raw, locales=["ak", "fr"])
        self.assertEqual(
            model.continent.name,
            None,
            "continent name is undef (no Akan or French " "available)",
        )
        self.assertEqual(model.country.name, "États-Unis", "country name is in French")

    def test_unknown_locale(self) -> None:
        model = geoip2.models.Country(self.raw, locales=["aa"])
        self.assertEqual(
            model.continent.name, None, "continent name is undef (no Afar available)"
        )
        self.assertEqual(
            model.country.name, None, "country name is in None (no Afar available)"
        )

    def test_german(self) -> None:
        model = geoip2.models.Country(self.raw, locales=["de"])
        self.assertEqual(
            model.continent.name, "Nordamerika", "Correct german name for continent"
        )


if __name__ == "__main__":
    unittest.main()



"""
Models
======

These classes provide models for the data returned by the GeoIP2
web service and databases.

The only difference between the City and Insights model classes is which
fields in each record may be populated. See
https://dev.maxmind.com/geoip/docs/web-services?lang=en for more details.

"""
# pylint: disable=too-many-instance-attributes,too-few-public-methods
import ipaddress
from abc import ABCMeta
from typing import Any, cast, Dict, List, Optional, Union

import geoip2.records
from geoip2.mixins import SimpleEquality


class Country(SimpleEquality):
    """Model for the Country web service and Country database.

    This class provides the following attributes:

    .. attribute:: continent

      Continent object for the requested IP address.

      :type: :py:class:`geoip2.records.Continent`

    .. attribute:: country

      Country object for the requested IP address. This record represents the
      country where MaxMind believes the IP is located.

      :type: :py:class:`geoip2.records.Country`

    .. attribute:: maxmind

      Information related to your MaxMind account.

      :type: :py:class:`geoip2.records.MaxMind`

    .. attribute:: registered_country

      The registered country object for the requested IP address. This record
      represents the country where the ISP has registered a given IP block in
      and may differ from the user's country.

      :type: :py:class:`geoip2.records.Country`

    .. attribute:: represented_country

      Object for the country represented by the users of the IP address
      when that country is different than the country in ``country``. For
      instance, the country represented by an overseas military base.

      :type: :py:class:`geoip2.records.RepresentedCountry`

    .. attribute:: traits

      Object with the traits of the requested IP address.

      :type: :py:class:`geoip2.records.Traits`

    """

    continent: geoip2.records.Continent
    country: geoip2.records.Country
    maxmind: geoip2.records.MaxMind
    registered_country: geoip2.records.Country
    represented_country: geoip2.records.RepresentedCountry
    traits: geoip2.records.Traits

    def __init__(
        self, raw_response: Dict[str, Any], locales: Optional[List[str]] = None
    ) -> None:
        if locales is None:
            locales = ["en"]
        self._locales = locales
        self.continent = geoip2.records.Continent(
            locales, **raw_response.get("continent", {})
        )
        self.country = geoip2.records.Country(
            locales, **raw_response.get("country", {})
        )
        self.registered_country = geoip2.records.Country(
            locales, **raw_response.get("registered_country", {})
        )
        self.represented_country = geoip2.records.RepresentedCountry(
            locales, **raw_response.get("represented_country", {})
        )

        self.maxmind = geoip2.records.MaxMind(**raw_response.get("maxmind", {}))

        self.traits = geoip2.records.Traits(**raw_response.get("traits", {}))
        self.raw = raw_response

    def __repr__(self) -> str:
        return (
            f"{self.__module__}.{self.__class__.__name__}({self.raw}, {self._locales})"
        )


class City(Country):
    """Model for the City Plus web service and the City database.

    .. attribute:: city

      City object for the requested IP address.

      :type: :py:class:`geoip2.records.City`

    .. attribute:: continent

      Continent object for the requested IP address.

      :type: :py:class:`geoip2.records.Continent`

    .. attribute:: country

      Country object for the requested IP address. This record represents the
      country where MaxMind believes the IP is located.

      :type: :py:class:`geoip2.records.Country`

    .. attribute:: location

      Location object for the requested IP address.

      :type: :py:class:`geoip2.records.Location`

    .. attribute:: maxmind

      Information related to your MaxMind account.

      :type: :py:class:`geoip2.records.MaxMind`

    .. attribute:: postal

      Postal object for the requested IP address.

      :type: :py:class:`geoip2.records.Postal`

    .. attribute:: registered_country

      The registered country object for the requested IP address. This record
      represents the country where the ISP has registered a given IP block in
      and may differ from the user's country.

      :type: :py:class:`geoip2.records.Country`

    .. attribute:: represented_country

      Object for the country represented by the users of the IP address
      when that country is different than the country in ``country``. For
      instance, the country represented by an overseas military base.

      :type: :py:class:`geoip2.records.RepresentedCountry`

    .. attribute:: subdivisions

      Object (tuple) representing the subdivisions of the country to which
      the location of the requested IP address belongs.

      :type: :py:class:`geoip2.records.Subdivisions`

    .. attribute:: traits

      Object with the traits of the requested IP address.

      :type: :py:class:`geoip2.records.Traits`

    """

    city: geoip2.records.City
    location: geoip2.records.Location
    postal: geoip2.records.Postal
    subdivisions: geoip2.records.Subdivisions

    def __init__(
        self, raw_response: Dict[str, Any], locales: Optional[List[str]] = None
    ) -> None:
        super().__init__(raw_response, locales)
        self.city = geoip2.records.City(locales, **raw_response.get("city", {}))
        self.location = geoip2.records.Location(**raw_response.get("location", {}))
        self.postal = geoip2.records.Postal(**raw_response.get("postal", {}))
        self.subdivisions = geoip2.records.Subdivisions(
            locales, *raw_response.get("subdivisions", [])
        )


class Insights(City):
    """Model for the GeoIP2 Insights web service.

    .. attribute:: city

      City object for the requested IP address.

      :type: :py:class:`geoip2.records.City`

    .. attribute:: continent

      Continent object for the requested IP address.

      :type: :py:class:`geoip2.records.Continent`

    .. attribute:: country

      Country object for the requested IP address. This record represents the
      country where MaxMind believes the IP is located.

      :type: :py:class:`geoip2.records.Country`

    .. attribute:: location

      Location object for the requested IP address.

    .. attribute:: maxmind

      Information related to your MaxMind account.

      :type: :py:class:`geoip2.records.MaxMind`

    .. attribute:: registered_country

      The registered country object for the requested IP address. This record
      represents the country where the ISP has registered a given IP block in
      and may differ from the user's country.

      :type: :py:class:`geoip2.records.Country`

    .. attribute:: represented_country

      Object for the country represented by the users of the IP address
      when that country is different than the country in ``country``. For
      instance, the country represented by an overseas military base.

      :type: :py:class:`geoip2.records.RepresentedCountry`

    .. attribute:: subdivisions

      Object (tuple) representing the subdivisions of the country to which
      the location of the requested IP address belongs.

      :type: :py:class:`geoip2.records.Subdivisions`

    .. attribute:: traits

      Object with the traits of the requested IP address.

      :type: :py:class:`geoip2.records.Traits`

    """


class Enterprise(City):
    """Model for the GeoIP2 Enterprise database.

    .. attribute:: city

      City object for the requested IP address.

      :type: :py:class:`geoip2.records.City`

    .. attribute:: continent

      Continent object for the requested IP address.

      :type: :py:class:`geoip2.records.Continent`

    .. attribute:: country

      Country object for the requested IP address. This record represents the
      country where MaxMind believes the IP is located.

      :type: :py:class:`geoip2.records.Country`

    .. attribute:: location

      Location object for the requested IP address.

    .. attribute:: maxmind

      Information related to your MaxMind account.

      :type: :py:class:`geoip2.records.MaxMind`

    .. attribute:: registered_country

      The registered country object for the requested IP address. This record
      represents the country where the ISP has registered a given IP block in
      and may differ from the user's country.

      :type: :py:class:`geoip2.records.Country`

    .. attribute:: represented_country

      Object for the country represented by the users of the IP address
      when that country is different than the country in ``country``. For
      instance, the country represented by an overseas military base.

      :type: :py:class:`geoip2.records.RepresentedCountry`

    .. attribute:: subdivisions

      Object (tuple) representing the subdivisions of the country to which
      the location of the requested IP address belongs.

      :type: :py:class:`geoip2.records.Subdivisions`

    .. attribute:: traits

      Object with the traits of the requested IP address.

      :type: :py:class:`geoip2.records.Traits`

    """


class SimpleModel(SimpleEquality, metaclass=ABCMeta):
    """Provides basic methods for non-location models"""

    raw: Dict[str, Union[bool, str, int]]
    ip_address: str
    _network: Optional[Union[ipaddress.IPv4Network, ipaddress.IPv6Network]]
    _prefix_len: int

    def __init__(self, raw: Dict[str, Union[bool, str, int]]) -> None:
        self.raw = raw
        self._network = None
        self._prefix_len = cast(int, raw.get("prefix_len"))
        self.ip_address = cast(str, raw.get("ip_address"))

    def __repr__(self) -> str:
        return f"{self.__module__}.{self.__class__.__name__}({self.raw})"

    @property
    def network(self) -> Optional[Union[ipaddress.IPv4Network, ipaddress.IPv6Network]]:
        """The network for the record"""
        # This code is duplicated for performance reasons
        network = self._network
        if network is not None:
            return network

        ip_address = self.ip_address
        prefix_len = self._prefix_len
        if ip_address is None or prefix_len is None:
            return None
        network = ipaddress.ip_network(f"{ip_address}/{prefix_len}", False)
        self._network = network
        return network


class AnonymousIP(SimpleModel):
    """Model class for the GeoIP2 Anonymous IP.

    This class provides the following attribute:

    .. attribute:: is_anonymous

      This is true if the IP address belongs to any sort of anonymous network.

      :type: bool

    .. attribute:: is_anonymous_vpn

      This is true if the IP address is registered to an anonymous VPN
      provider.

      If a VPN provider does not register subnets under names associated with
      them, we will likely only flag their IP ranges using the
      ``is_hosting_provider`` attribute.

      :type: bool

    .. attribute:: is_hosting_provider

      This is true if the IP address belongs to a hosting or VPN provider
      (see description of ``is_anonymous_vpn`` attribute).

      :type: bool

    .. attribute:: is_public_proxy

      This is true if the IP address belongs to a public proxy.

      :type: bool

    .. attribute:: is_residential_proxy

      This is true if the IP address is on a suspected anonymizing network
      and belongs to a residential ISP.

      :type: bool

    .. attribute:: is_tor_exit_node

      This is true if the IP address is a Tor exit node.

      :type: bool

    .. attribute:: ip_address

      The IP address used in the lookup.

      :type: str

    .. attribute:: network

      The network associated with the record. In particular, this is the
      largest network where all of the fields besides ip_address have the same
      value.

      :type: ipaddress.IPv4Network or ipaddress.IPv6Network
    """

    is_anonymous: bool
    is_anonymous_vpn: bool
    is_hosting_provider: bool
    is_public_proxy: bool
    is_residential_proxy: bool
    is_tor_exit_node: bool

    def __init__(self, raw: Dict[str, bool]) -> None:
        super().__init__(raw)  # type: ignore
        self.is_anonymous = raw.get("is_anonymous", False)
        self.is_anonymous_vpn = raw.get("is_anonymous_vpn", False)
        self.is_hosting_provider = raw.get("is_hosting_provider", False)
        self.is_public_proxy = raw.get("is_public_proxy", False)
        self.is_residential_proxy = raw.get("is_residential_proxy", False)
        self.is_tor_exit_node = raw.get("is_tor_exit_node", False)


class ASN(SimpleModel):
    """Model class for the GeoLite2 ASN.

    This class provides the following attribute:

    .. attribute:: autonomous_system_number

      The autonomous system number associated with the IP address.

      :type: int

    .. attribute:: autonomous_system_organization

      The organization associated with the registered autonomous system number
      for the IP address.

      :type: str

    .. attribute:: ip_address

      The IP address used in the lookup.

      :type: str

    .. attribute:: network

      The network associated with the record. In particular, this is the
      largest network where all of the fields besides ip_address have the same
      value.

      :type: ipaddress.IPv4Network or ipaddress.IPv6Network
    """

    autonomous_system_number: Optional[int]
    autonomous_system_organization: Optional[str]

    # pylint:disable=too-many-arguments
    def __init__(self, raw: Dict[str, Union[str, int]]) -> None:
        super().__init__(raw)
        self.autonomous_system_number = cast(
            Optional[int], raw.get("autonomous_system_number")
        )
        self.autonomous_system_organization = cast(
            Optional[str], raw.get("autonomous_system_organization")
        )


class ConnectionType(SimpleModel):
    """Model class for the GeoIP2 Connection-Type.

    This class provides the following attribute:

    .. attribute:: connection_type

      The connection type may take the following values:

      - Dialup
      - Cable/DSL
      - Corporate
      - Cellular
      - Satellite

      Additional values may be added in the future.

      :type: str

    .. attribute:: ip_address

      The IP address used in the lookup.

      :type: str

    .. attribute:: network

      The network associated with the record. In particular, this is the
      largest network where all of the fields besides ip_address have the same
      value.

      :type: ipaddress.IPv4Network or ipaddress.IPv6Network
    """

    connection_type: Optional[str]

    def __init__(self, raw: Dict[str, Union[str, int]]) -> None:
        super().__init__(raw)
        self.connection_type = cast(Optional[str], raw.get("connection_type"))


class Domain(SimpleModel):
    """Model class for the GeoIP2 Domain.

    This class provides the following attribute:

    .. attribute:: domain

      The domain associated with the IP address.

      :type: str

    .. attribute:: ip_address

      The IP address used in the lookup.

      :type: str

    .. attribute:: network

      The network associated with the record. In particular, this is the
      largest network where all of the fields besides ip_address have the same
      value.

      :type: ipaddress.IPv4Network or ipaddress.IPv6Network

    """

    domain: Optional[str]

    def __init__(self, raw: Dict[str, Union[str, int]]) -> None:
        super().__init__(raw)
        self.domain = cast(Optional[str], raw.get("domain"))


class ISP(ASN):
    """Model class for the GeoIP2 ISP.

    This class provides the following attribute:

    .. attribute:: autonomous_system_number

      The autonomous system number associated with the IP address.

      :type: int

    .. attribute:: autonomous_system_organization

      The organization associated with the registered autonomous system number
      for the IP address.

      :type: str

    .. attribute:: isp

      The name of the ISP associated with the IP address.

      :type: str

    .. attribute: mobile_country_code

      The `mobile country code (MCC)
      <https://en.wikipedia.org/wiki/Mobile_country_code>`_ associated with the
      IP address and ISP.

      :type: str

    .. attribute: mobile_network_code

      The `mobile network code (MNC)
      <https://en.wikipedia.org/wiki/Mobile_country_code>`_ associated with the
      IP address and ISP.

      :type: str

    .. attribute:: organization

      The name of the organization associated with the IP address.

      :type: str

    .. attribute:: ip_address

      The IP address used in the lookup.

      :type: str

    .. attribute:: network

      The network associated with the record. In particular, this is the
      largest network where all of the fields besides ip_address have the same
      value.

      :type: ipaddress.IPv4Network or ipaddress.IPv6Network
    """

    isp: Optional[str]
    mobile_country_code: Optional[str]
    mobile_network_code: Optional[str]
    organization: Optional[str]

    # pylint:disable=too-many-arguments
    def __init__(self, raw: Dict[str, Union[str, int]]) -> None:
        super().__init__(raw)
        self.isp = cast(Optional[str], raw.get("isp"))
        self.mobile_country_code = cast(Optional[str], raw.get("mobile_country_code"))
        self.mobile_network_code = cast(Optional[str], raw.get("mobile_network_code"))
        self.organization = cast(Optional[str], raw.get("organization"))



"""
======================
GeoIP2 Database Reader
======================

"""
import inspect
import os
from typing import Any, AnyStr, cast, IO, List, Optional, Type, Union

import maxminddb

from maxminddb import (
    MODE_AUTO,
    MODE_MMAP,
    MODE_MMAP_EXT,
    MODE_FILE,
    MODE_MEMORY,
    MODE_FD,
)

import geoip2
import geoip2.models
import geoip2.errors
from geoip2.types import IPAddress
from geoip2.models import (
    ASN,
    AnonymousIP,
    City,
    ConnectionType,
    Country,
    Domain,
    Enterprise,
    ISP,
)

__all__ = [
    "MODE_AUTO",
    "MODE_MMAP",
    "MODE_MMAP_EXT",
    "MODE_FILE",
    "MODE_MEMORY",
    "MODE_FD",
    "Reader",
]


class Reader:
    """GeoIP2 database Reader object.

    Instances of this class provide a reader for the GeoIP2 database format.
    IP addresses can be looked up using the ``country`` and ``city`` methods.

    The basic API for this class is the same for every database. First, you
    create a reader object, specifying a file name or file descriptor.
    You then call the method corresponding to the specific database, passing
    it the IP address you want to look up.

    If the request succeeds, the method call will return a model class for the
    method you called. This model in turn contains multiple record classes,
    each of which represents part of the data returned by the database. If the
    database does not contain the requested information, the attributes on the
    record class will have a ``None`` value.

    If the address is not in the database, an
    ``geoip2.errors.AddressNotFoundError`` exception will be thrown. If the
    database is corrupt or invalid, a ``maxminddb.InvalidDatabaseError`` will
    be thrown.
    """

    def __init__(
        self,
        fileish: Union[AnyStr, int, os.PathLike, IO],
        locales: Optional[List[str]] = None,
        mode: int = MODE_AUTO,
    ) -> None:
        """Create GeoIP2 Reader.

        :param fileish: A path to the GeoIP2 database or an existing file
          descriptor pointing to the database. Note that a file descriptor
          is only valid when mode is MODE_FD.
        :param locales: This is list of locale codes. This argument will be
          passed on to record classes to use when their name properties are
          called. The default value is ['en'].

          The order of the locales is significant. When a record class has
          multiple names (country, city, etc.), its name property will return
          the name in the first locale that has one.

          Note that the only locale which is always present in the GeoIP2
          data is "en". If you do not include this locale, the name property
          may end up returning None even when the record has an English name.

          Currently, the valid locale codes are:

          * de -- German
          * en -- English names may still include accented characters if that
            is the accepted spelling in English. In other words, English does
            not mean ASCII.
          * es -- Spanish
          * fr -- French
          * ja -- Japanese
          * pt-BR -- Brazilian Portuguese
          * ru -- Russian
          * zh-CN -- Simplified Chinese.
        :param mode: The mode to open the database with. Valid mode are:
          * MODE_MMAP_EXT - use the C extension with memory map.
          * MODE_MMAP - read from memory map. Pure Python.
          * MODE_FILE - read database as standard file. Pure Python.
          * MODE_MEMORY - load database into memory. Pure Python.
          * MODE_FD - the param passed via fileish is a file descriptor, not a
             path. This mode implies MODE_MEMORY. Pure Python.
          * MODE_AUTO - try MODE_MMAP_EXT, MODE_MMAP, MODE_FILE in that order.
             Default.

        """
        if locales is None:
            locales = ["en"]
        self._db_reader = maxminddb.open_database(fileish, mode)
        self._db_type = self._db_reader.metadata().database_type
        self._locales = locales

    def __enter__(self) -> "Reader":
        return self

    def __exit__(self, exc_type: None, exc_value: None, traceback: None) -> None:
        self.close()

    def country(self, ip_address: IPAddress) -> Country:
        """Get the Country object for the IP address.

        :param ip_address: IPv4 or IPv6 address as a string.

        :returns: :py:class:`geoip2.models.Country` object

        """

        return cast(
            Country, self._model_for(geoip2.models.Country, "Country", ip_address)
        )

    def city(self, ip_address: IPAddress) -> City:
        """Get the City object for the IP address.

        :param ip_address: IPv4 or IPv6 address as a string.

        :returns: :py:class:`geoip2.models.City` object

        """
        return cast(City, self._model_for(geoip2.models.City, "City", ip_address))

    def anonymous_ip(self, ip_address: IPAddress) -> AnonymousIP:
        """Get the AnonymousIP object for the IP address.

        :param ip_address: IPv4 or IPv6 address as a string.

        :returns: :py:class:`geoip2.models.AnonymousIP` object

        """
        return cast(
            AnonymousIP,
            self._flat_model_for(
                geoip2.models.AnonymousIP, "GeoIP2-Anonymous-IP", ip_address
            ),
        )

    def asn(self, ip_address: IPAddress) -> ASN:
        """Get the ASN object for the IP address.

        :param ip_address: IPv4 or IPv6 address as a string.

        :returns: :py:class:`geoip2.models.ASN` object

        """
        return cast(
            ASN, self._flat_model_for(geoip2.models.ASN, "GeoLite2-ASN", ip_address)
        )

    def connection_type(self, ip_address: IPAddress) -> ConnectionType:
        """Get the ConnectionType object for the IP address.

        :param ip_address: IPv4 or IPv6 address as a string.

        :returns: :py:class:`geoip2.models.ConnectionType` object

        """
        return cast(
            ConnectionType,
            self._flat_model_for(
                geoip2.models.ConnectionType, "GeoIP2-Connection-Type", ip_address
            ),
        )

    def domain(self, ip_address: IPAddress) -> Domain:
        """Get the Domain object for the IP address.

        :param ip_address: IPv4 or IPv6 address as a string.

        :returns: :py:class:`geoip2.models.Domain` object

        """
        return cast(
            Domain,
            self._flat_model_for(geoip2.models.Domain, "GeoIP2-Domain", ip_address),
        )

    def enterprise(self, ip_address: IPAddress) -> Enterprise:
        """Get the Enterprise object for the IP address.

        :param ip_address: IPv4 or IPv6 address as a string.

        :returns: :py:class:`geoip2.models.Enterprise` object

        """
        return cast(
            Enterprise,
            self._model_for(geoip2.models.Enterprise, "Enterprise", ip_address),
        )

    def isp(self, ip_address: IPAddress) -> ISP:
        """Get the ISP object for the IP address.

        :param ip_address: IPv4 or IPv6 address as a string.

        :returns: :py:class:`geoip2.models.ISP` object

        """
        return cast(
            ISP, self._flat_model_for(geoip2.models.ISP, "GeoIP2-ISP", ip_address)
        )

    def _get(self, database_type: str, ip_address: IPAddress) -> Any:
        if database_type not in self._db_type:
            caller = inspect.stack()[2][3]
            raise TypeError(
                f"The {caller} method cannot be used with the {self._db_type} database",
            )
        (record, prefix_len) = self._db_reader.get_with_prefix_len(ip_address)
        if record is None:
            raise geoip2.errors.AddressNotFoundError(
                f"The address {ip_address} is not in the database.",
                str(ip_address),
                prefix_len,
            )
        return record, prefix_len

    def _model_for(
        self,
        model_class: Union[Type[Country], Type[Enterprise], Type[City]],
        types: str,
        ip_address: IPAddress,
    ) -> Union[Country, Enterprise, City]:
        (record, prefix_len) = self._get(types, ip_address)
        traits = record.setdefault("traits", {})
        traits["ip_address"] = ip_address
        traits["prefix_len"] = prefix_len
        return model_class(record, locales=self._locales)

    def _flat_model_for(
        self,
        model_class: Union[
            Type[Domain], Type[ISP], Type[ConnectionType], Type[ASN], Type[AnonymousIP]
        ],
        types: str,
        ip_address: IPAddress,
    ) -> Union[ConnectionType, ISP, AnonymousIP, Domain, ASN]:
        (record, prefix_len) = self._get(types, ip_address)
        record["ip_address"] = ip_address
        record["prefix_len"] = prefix_len
        return model_class(record)

    def metadata(
        self,
    ) -> maxminddb.reader.Metadata:
        """The metadata for the open database.

        :returns: :py:class:`maxminddb.reader.Metadata` object
        """
        return self._db_reader.metadata()

    def close(self) -> None:
        """Closes the GeoIP2 database."""

        self._db_reader.close()



"""This package contains utility mixins"""
# pylint: disable=too-few-public-methods
from abc import ABCMeta
from typing import Any


class SimpleEquality(metaclass=ABCMeta):
    """Naive __dict__ equality mixin"""

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)
